#include "raysphere.h"
#include "noises.h"
#include "filterwidth.h"

/*
An implemetation of a "BSSRDF Lighting Model" by Mohamed Sobhy@NCCA 2003
*/


/*-------------------------------------------------------------------------------------------------------------------*/
float volumedensity (point Pobj; float radius, noisefreq, stepsize;)
{
    float density = 0.5 + 0.5 * fBm(Pobj*noisefreq, stepsize*noisefreq, 7, 2, 0.65);
    density = pow(clamp(density,0,1), 7);
    return density;
}

/*-------------------------------------------------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------------------------------------------------*/
/*
A beam of light pasing through a medium will have its intensity reduced by e^-extcoef*s
"extcoef" = extictionCoefficient
"s" = distance throught the media
*/

float intensityReduction(float extCoef, s;)
{
float A = (-1)*extCoef*s; 
return exp(A);
}

/*-------------------------------------------------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------------------------------------------------*/

/*
The phase function describes the distribution of scattered light in a participating media.
----- HENYEY-GREENSTEIN PHASE FUCNTION ----

The phase function depends on the angle, "thita", between the incoming light and the scattered light.
if g=0 then p(thita)= 1/4PI;
else
p(thita) = (1-g^2) / 4PI(1+g^2 - 2gcos(thita))^1.5

Where "g" E[-1,1], is an asymmetry paramter equal to the average cosine of the scattered directions. 
Positive "g" gives forward scattering. Negative "g" gives backward scattering.
*/

float HGPhase(vector RefRay, lightRay; float g;)
{ 
	float B = 4*PI;
	float phase;
	if (g==0)
	{
		phase = 1/B;
	}
		else
	{
		float CosThita = RefRay.lightRay; 
		float A= (1 - pow(g, 2));
	
		float C= (1+(pow(g, 2)) - 2*g*CosThita);
		phase = A / B * pow(C, 1.5);
	}
	
	return phase;
}

/*-------------------------------------------------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------------------------------------------------*/

/* calculates refracted ray */
vector refRay(vector II; normal NN; float eta; output float Kr, Kt;) 
{
	vector R;	/* reflection direction */
	vector T;	/* transmitted direction */
	fresnel(II, NN, eta, Kr, Kt, R, T);  	/* fresnel function */
	Kr = smoothstep(0, 0.5, Kr);		/* a smooth stepped scaling factor for refrected light */
	Kt = 1.0 - Kr;		/* scaling factor for refrected light */
	
	return normalize(T);	/* refracted ray */
}
/*-------------------------------------------------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------------------------------------------------*/

color SubSurfaceCalc (point currentP; vector RefractedRay; float g, eta, totalStep, extcoef, radius, dSamplePoints, noisefreq, DensityScaler;)
{
    color translucency=0;
    float HGPhase=0;
    float allDist=0;
    color tcolor = 0;
    float denStepSize = 0;
	 float distanceTraveled = 0;
 	 point denSubsurfacePoint = 0;
	 point PObjectden = 0;
    float lastDensity = 0;
	 float Density = 0;
	 float dss = 0;
	 float DistanceToLight = 0;


	 illuminance (currentP) {
	
	extern color Cl;
	extern vector L;
	vector LN;
	
	LN = normalize(L);
	/* code to calculate the distance light has traveled */
	
	point  Pobj = transform ("object", currentP);
	vector Iobj = normalize (vtransform ("object", LN));
	float t0, t1;
	float hits = raysphere (Pobj, Iobj, radius, 1.0e-4, t0, t1);
	float DistanceToLight = (hits > 0) ? t0 : 0;  /* distance to march */

	denStepSize = DistanceToLight/dSamplePoints;
	
	float distanceTraveled = random()*denStepSize;	/* Integrate forwards from the start point */
	float dss = min (denStepSize, DistanceToLight-distanceTraveled);	/* Calculate a reasonable step size */
	
	while (distanceTraveled < DistanceToLight) 
	{
	  distanceTraveled += dss;
	  denSubsurfacePoint = currentP + (distanceTraveled*LN);
	  
	  PObjectden  = transform("object", denSubsurfacePoint);
     Density += volumedensity (PObjectden , radius, noisefreq, denStepSize);
	} 
	Density = Density * DensityScaler * denStepSize;

	/*works out real distance traveled*/
 	allDist = DistanceToLight + totalStep;
	float intensityScaler = intensityReduction(extcoef, totalStep); 
	
	/* HENYEY-GREENSTEIN PHASE FUCNTION */
	HGPhase+= HGPhase(RefractedRay, LN, g);
   tcolor += Cl; 
	translucency+=HGPhase*intensityScaler*tcolor*Density;
	
	}
    return translucency;
}
/*-------------------------------------------------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------------------------------------------------*/
surface
subsurfaceScatter(float Ka = 0.127, Kd = 0.8, Ks = 0.5;
		color SurfaceColor = color (0, 0, 1.0);
		color SpecularColor = color (1.0, 1.0, 1.0);
		float roughness = 0.5;
		float maxDepth = 1.0;    /* normalized value, 0 no depth, 1 full depth*/ 
		float samplePoints = 4;	 /* number of samples  */
		float densitySamplePoints = 1;
		float extinctionCoef = 1;	/* extinction coef = absorbtion coef + scatter coef */
		float g = 0.5;
		float eta = 1.5;
		float radius = 3.0;
		float noisefreq = 0.4;
		float DensityScaler = 10.0;
		)
{
    /* inilizing renderman variables */
    extern vector I;
    extern point P;
    point subsurfacePoint;
    vector V = normalize(-I);
    vector IN = normalize(I);
    vector NN = normalize(N);
    normal Nf = faceforward(normalize(N), I);
    float Kr, Kt;
    point Ph;	/* position of point hit*/ 
    vector Nh;  /* normal of point hit */
    color subLevelC=0;
    float stepsize = 0;

   	/* calculate refracted ray */
   	vector r = refRay(IN, NN, 1/eta, Kr, Kt);
   	vector RefractedRay= normalize(r);
   	

	point  Pobj = transform ("object", P);
	vector Iobj = normalize (vtransform ("object", RefractedRay));
	float t0, t1;
	float hits = raysphere (Pobj, Iobj, radius, 1.0e-4, t0, t1);
	float maxDist = (hits > 0) ? t0 : 0;  /* distance to march */
	
	if ( maxDist > 0 )
	{
	stepsize = maxDist/samplePoints;
	
	float distanceTraveled = random()*stepsize;	/* Integrate forwards from the start point */
	float ss = min (stepsize, maxDist-distanceTraveled);	/* Calculate a reasonable step size */
	
	subsurfacePoint = P + (distanceTraveled*RefractedRay);
	
		while (distanceTraveled <= maxDist) 
			{
	  			distanceTraveled += ss;
	  			subsurfacePoint = P + (distanceTraveled*RefractedRay);
				subLevelC += ss * SubSurfaceCalc(subsurfacePoint, RefractedRay, g, eta, distanceTraveled, extinctionCoef, radius, densitySamplePoints, noisefreq, DensityScaler);
			}
 	}

	 Ci = SurfaceColor * (Ka*ambient()+Kd*diffuse(Nf)) + Ks*SpecularColor*specular(NN,V,roughness)+subLevelC;
	 	 	Oi = Os;
	 	 	Ci*=Oi;
    
    
}
/*-------------------------------------------------------------------------------------------------------------------*/



